class GameWorld { 

  //measurments of the GameWorld
  final float worldWidthMeters = 4;
  final float worldHeightMeters = 2.25;

  final float HOOKES_CONSTANT= 0.05;
  float goodBadBubbleRatio=0.4;

  //variables for lives
  int lives=3;
  int maxLives=3;

  //variables for bubbles
  int numberOfBubbles=14;
  int bubbleScore=0;
  int bubbleCounter=0; //counts number Of bubbles
  int bubbleRatioCounter=1;

  //starting positions for unicorns
  PVector startPosP1=new PVector(worldWidthMeters/2, worldHeightMeters/2);
  PVector startPosP2=new PVector(worldWidthMeters/3, worldHeightMeters/2);

  boolean[] keys = new boolean[255];

  // Controls to move unicorns 
  int wPressed = 'w', aPressed = 'a', sPressed = 's', dPressed = 'd';
  int noPressed;
  int upPressed = UP, downPressed = DOWN, leftPressed = LEFT, rightPressed = RIGHT;

  //variables for save respawn
  float saveRespawnTime=0;
  boolean saveRespawn=false;

  //variables for RainbowMode
  boolean rainbowMode=false;
  int rainbowSpeed=8;
  float rainbowTime=0; 
  int pulseCounter=0;//how the rainbowMeter pulses

  //variables for Glitter
  int glitterScore=0; //current glitter score
  int maxGlitterScore=5; //how many glitter you need to get in the rainbowMode
  int glitterCounter=0; //how many glitters are in the game
  int maxAmountOfGlitter=18;//how many glitters should be in the game at all times

  //current game height (score)
  int gameHeight=0;

  float updateSpeed=1;

  float gamePausedTime=0;

  float speed=0;


  java.util.LinkedList<GameObject> gameObjects;   // list with all game objects

  //Spring force
  Spring s1;
  Spring s2;
  //2 Players
  Unicorn player1;
  Unicorn player2;
  //sweets
  Sweet[] sweets;
  //Background
  Background background;


  GameWorld() {

    // creating a list containing all gameobjects
    gameObjects = new java.util.LinkedList<GameObject>();

    // construct a default scene by creating gameObjects
    background = new Background(this, new PVector(0, 0));

    player1 = new Unicorn(this, startPosP1.x, startPosP1.y, player1Icon);
    player2 = new Unicorn(this, startPosP2.x, startPosP2.y, player2Icon);

//Spring between players, length: 0.1 meter
    s1 = new Spring(this, player1, player2, 0.15);
    s2 = new Spring(this, player2, player1, 0.15);

    //adding background before everything else
    gameObjects.add(background);


    //spawn random bubbles without overlapping
    for (int i=0; i<numberOfBubbles; i++) {
      println(i);

      float x = random(worldWidthMeters);      
      float y = random(worldHeightMeters);
      float radius=random(0.2, 0.3);

      //try to place Bubbles and stop if not possible
    attempts: 
      for (int attempts=0; attempts<20; attempts++) {
        for (GameObject o : gameObjects) {
          if (o instanceof Bubble) {
            Bubble bubble= (Bubble) o;
            //calculate distance between Bubble and Bubble and between Player and Bubble
            double bubbleDistance = Math.sqrt((bubble.s.x-x)*(bubble.s.x-x)+(bubble.s.y-y)*(bubble.s.y-y));
            double player1Distance= Math.sqrt((startPosP1.x-x)*(startPosP1.x-x)+(startPosP1.y-y)*(startPosP1.y-y));
            double player2Distance= Math.sqrt((startPosP2.x-x)*(startPosP2.x-x)+(startPosP2.y-y)*(startPosP2.y-y));

            if (bubbleDistance<=(radius+bubble.radius+0.05)||player1Distance<=(radius+player1.radius+0.25)||player2Distance<=(radius+player2.radius+0.25)) {
              //new Position, if Bubble overlaps with player/ other Bubbles
              x = random(worldWidthMeters);      
              y=  random(worldHeightMeters);
              continue attempts;
            }
          }
        }
        if (i<numberOfBubbles*goodBadBubbleRatio) {
          gameObjects.add(new BadBubble(this, x, y, radius, false));
          bubbleCounter++;
          break;
        } else {
          Sweet sweet=new Sweet(this, x+radius/2, y-radius/2, 0.3);  
          gameObjects.add(new GoodBubble(this, x, y, radius, false, sweet));
          gameObjects.add(sweet);
          bubbleCounter++;
          break;
        }
      }
    }

    //spawn random glitter
    for (int i = 0; i < maxAmountOfGlitter; ++i) {
      float x = random(worldWidthMeters);      
      float y = random(worldHeightMeters);
      float radius= 0.15;
      PImage glitterImage = glitter;
      gameObjects.add(new Glitter(this, x, y, radius, glitterImage));
      glitterCounter++;
    }

    //add the other GameObjects to the list
    gameObjects.add(s1);
    gameObjects.add(s2);

    gameObjects.add(player1);
    gameObjects.add(player2);

    //top boundary
    gameObjects.add(new WorldBoundary(this, new PVector(worldWidthMeters, 0), new PVector(0, 0)));
    //right boundary
    gameObjects.add(new WorldBoundary(this, new PVector(worldWidthMeters-0.1, worldHeightMeters), new PVector(worldWidthMeters-0.1, 0)));
  }

  void animate(float deltaTSeconds) {

    s1.update(deltaTSeconds);
    s2.update(deltaTSeconds);


    for (GameObject o : gameObjects) {
      o.updateA(deltaTSeconds);
    }

    for (GameObject o : gameObjects) {
      o.updateV(deltaTSeconds);
    }

    for (GameObject o : gameObjects) {
      o.updateS(deltaTSeconds);
    }
    for (GameObject o : gameObjects) {
      o.handleCollisions(deltaTSeconds);
    }

    destroyBubbleOutOfBounds();
    handleUnicornOutOfScreen();
    addBubbleRatio();
  }

  void draw() { 
    checkForGameOver();

    // draw all foreground objects. 
    for (GameObject o : gameObjects) {
      o.draw();
    }
  }


  /* void cheatRainbowMode() { //debug cheat
    if (rainbowMode == false) glitterScore=maxGlitterScore;
    else if (rainbowMode == true) setRainbowModeFalse(); 
  } */

  //deactivates RaibowMode
  void setRainbowModeFalse() {
    rainbowMode=false;
    glitterScore=0;
    pulseCounter=0;
    unicornMode.pause(); //AUDIO
    ingame.unmute(); //AUDIO
    nightSky_ingame.unmute(); //AUDIO
  }
  //activates RainbowMode
  public void setRainbowModeTrue() {
    rainbowMode=true;
    gamePausedTime=0;
    rainbowMeterDing.pause();//AUDIO
    rainbowMeterDing.rewind();//AUDIO
    ingame.mute(); //AUDIO
    nightSky_ingame.mute(); //AUDIO
    unicornMode.play(); //AUDIO
    unicornMode.loop(); //AUDIO
    unicornMode.rewind(); //AUDIO
    rainbowTime=millis();
  }

  //player loses Live by moving out of screen
  void handleUnicornOutOfScreen() {

    if (player1.s.x+player1.radius-0.09<0||player1.s.y-player1.radius+0.09>worldHeightMeters
      ||player2.s.x+player2.radius-0.09<0||player2.s.y-player2.radius+0.09>worldHeightMeters) {//is player out of screen?
      lives--;
      //println(lives); //Debug
      if (rainbowMode==true) {
        setRainbowModeFalse();//deactovate rainbowMode
      }
      lostHeart.play();//AUDIO
      lostHeart.rewind();//AUDIO


      //reset PlayerPos to startPos
      player1.s=startPosP1.copy();
      player2.s=startPosP2.copy();

      //if Player respawns they can't be hurt for 3sec
      saveRespawn=true;
      saveRespawnTime = millis();
    }
  }

  void checkForGameOver() {
    if (lives<=0) {
      gameOver();
    }
  }

  void gameOver() {      
    //saves highscore and last three scores
    if (indexAllScores < allScores.length) {
      finalScore = gameHeight;
      for (int i=scoreArray.length-2; i>=0; i--) {
        scoreArray[i+1]=scoreArray[i];
      }
      scoreArray[0] = finalScore;
      allScores[indexAllScores]  = finalScore;
      indexAllScores++;
      if (highScore < finalScore) {
        highScore = finalScore;
      }
    } else {
      indexAllScores = 0;
      finalScore = gameHeight;
      for (int i=scoreArray.length-2; i>=0; i--) {
        scoreArray[i+1]=scoreArray[i];
      }
      scoreArray[0] = finalScore;
      allScores[indexAllScores]  = finalScore;
      indexAllScores++;
      if (highScore < finalScore) {
        highScore = finalScore;
      }
    }
    gameMode=new GameEndsMode();
  }


  void moveUnicorn(int pressedKey) {
    if (pressedKey =='w' || pressedKey == 'W') {
      player2.isMoving=true;
      player2.move(wPressed);
    }
    if (pressedKey == 's' || pressedKey == 'S') {
      player2.isMoving=true;
      player2.move(sPressed);
    }

    if (pressedKey == 'a'|| pressedKey == 'A') {
      player2.isMoving=true;
      player2.move(aPressed);
    }

    if (pressedKey == 'd'|| pressedKey == 'D') {
      player2.isMoving=true;
      player2.move(dPressed);
    }

    if (pressedKey == noPressed) {
      player2.isMoving = false;
      player2.move(noPressed);
    }
    if (pressedKey == UP) {
      player1.isMoving=true;
      player1.move(upPressed);
    }
    if (pressedKey == DOWN) {
      player1.isMoving=true;
      player1.move(downPressed);
    }

    if (pressedKey == LEFT) {
      player1.isMoving=true;
      player1.move(leftPressed);
    }

    if (pressedKey == RIGHT) {
      player1.isMoving=true;
      player1.move(rightPressed);
    }

    if (pressedKey == noPressed) {
      player1.isMoving = false;
      player1.move(noPressed);
    }
  }

  //Destroy Bubbles, Sweets, Glitter that are no longer in the map
  void destroyBubbleOutOfBounds() {
    java.util.LinkedList<GameObject> objectsToRemove = new java.util.LinkedList<GameObject>();
    for (GameObject o : gameObjects) {
      if (o instanceof SphericalGameObject) {
        SphericalGameObject sgo=(SphericalGameObject)o;

        if (sgo instanceof Bubble) {//is bubble?
          Bubble b=(Bubble)sgo;
          if (b.s.y-b.radius>worldHeightMeters||b.s.x-b.radius>worldWidthMeters||b.s.y+b.radius<0||b.s.x+b.radius<0) {// is out of screen? ->remove Bubble
            objectsToRemove.add(b);
            bubbleCounter--;
            if (b instanceof GoodBubble) { //is Good Bubble -> also remove Sweet
              GoodBubble gb=(GoodBubble)b;
              objectsToRemove.add(gb.sweet);
            }
          }
          if (b.popped==true) { //is bubble popped ->remove Bubble    
            objectsToRemove.add(b);
            if (b instanceof GoodBubble&&lives==3) { //full lives-> also remove Sweet
              GoodBubble gb=(GoodBubble)b;
              objectsToRemove.add(gb.sweet);
            }
          }
        }
        if (sgo instanceof Sweet) {//is Sweet
          Sweet s=(Sweet)sgo;
          if (s.withoutBubble==true) {//sweet without Bubble?
            //if sweet reaches heart:
            if (s.s.y<0.22&&s.s.x<0.22) { 
              objectsToRemove.add(s);
              if (lives < 3) { //if lives are not full
                bubbleScore++;
                if (bubbleScore<7) {
                  heartFragment.play(); //AUDIO
                  heartFragment.rewind(); //AUDIO
                  //score counts popped bubbles
                } else if (bubbleScore >= 7) {//regain a live after eating enough sweets
                  lives++;
                  bubbleScore = 0;
                  yay.play(); //AUDIO
                  yay.rewind(); //AUDIO
                }
              }
            }
          }
        }
        if (sgo instanceof Glitter) {//is Glitter?
          Glitter g = (Glitter)sgo;
          if (g.popped == true) {//remove if Glitter is popped
            objectsToRemove.add(g);
          }
        }
      }
    }
    gameObjects.removeAll(objectsToRemove);//remove all Objects to remove

    if (bubbleCounter<numberOfBubbles) {//spawns new Bubbles, if there are less Bubbles left than numberOfBubbles
      spawnNewBubble();
    }
    if (glitterCounter<maxAmountOfGlitter) {//spawns new Glitter, if there is less Glitter left than maxAmountOfGlitter
      spawnNewGlitter();
    }
  }

  //increase Bubble Ratio when player collects more sweets to increase the difficulty
  void addBubbleRatio() {

    if (bubbleRatioCounter*400<gameHeight && goodBadBubbleRatio<0.64) {
      goodBadBubbleRatio+=0.03;
      if (numberOfBubbles>8) {//decrease numberOfBubbles to keep a chance to stay alive
        numberOfBubbles--; 
        //println(numberOfBubbles);//debug
      }
      bubbleRatioCounter++;
    }
  }

//spawns new Glitter
  void spawnNewGlitter() {
    Glitter g;
    float x = random(worldWidthMeters);      
    float y = random(worldHeightMeters);
    PImage glitterImage = glitter;
    float radius= 0.15;
  attempts: 
    for (int attempts=0; attempts<20; attempts++) {
      for (GameObject o : gameObjects) {
        if (o instanceof Glitter) {
          Glitter glitter= (Glitter) o;
          double glitterDistance = Math.sqrt((glitter.s.x-x)*(glitter.s.x-x)+(glitter.s.y-y)*(glitter.s.y-y));
          double player1Distance= Math.sqrt((player1.s.x-x)*(player1.s.x-x)+(player1.s.y-y)*(player1.s.y-y));
          double player2Distance= Math.sqrt((player2.s.x-x)*(player2.s.x-x)+(player2.s.y-y)*(player2.s.y-y));

          if (glitterDistance<=(radius+glitter.radius+0.05)||player1Distance<=(radius+player1.radius+0.2)||player2Distance<=(radius+player2.radius+0.42)) {
            x = random(worldWidthMeters);      
            y=  random(worldHeightMeters);
            continue attempts;
          }
        }
      }
      g=new Glitter(this, x, y, radius, glitterImage);
      gameObjects.add(g);
      glitterCounter++;
      break;
    }
  }

  //spawn new Bubble
  void spawnNewBubble() {

    Bubble b;
    float random=random(0, 1);

    float x = random(worldWidthMeters);      
    float y = random(worldHeightMeters);
    float radius=random(0.2, 0.3);
  attempts: 
    for (int attempts=0; attempts<20; attempts++) {
      for (GameObject o : gameObjects) {
        if (o instanceof Bubble) {
          Bubble bubble= (Bubble) o;
          double bubbleDistance = Math.sqrt((bubble.s.x-x)*(bubble.s.x-x)+(bubble.s.y-y)*(bubble.s.y-y));
          double player1Distance= Math.sqrt((player1.s.x-x)*(player1.s.x-x)+(player1.s.y-y)*(player1.s.y-y));
          double player2Distance= Math.sqrt((player2.s.x-x)*(player2.s.x-x)+(player2.s.y-y)*(player2.s.y-y));

          if (bubbleDistance<=(radius+bubble.radius+0.1)||player1Distance<=(radius+player1.radius+0.4)||player2Distance<=(radius+player2.radius+0.4)) {
            x = random(worldWidthMeters);      
            y=  random(worldHeightMeters);
            continue attempts;
          }
        }
      }
      if (random<goodBadBubbleRatio) {
        b= new BadBubble(this, x, y, radius, false);
        gameObjects.add(b);
        bubbleCounter++;
        break;
      } else {
        b= new GoodBubble(this, x, y, radius, false, new Sweet(this, x, y, 0.3));
        gameObjects.add(b);
        bubbleCounter++;
        if (b instanceof GoodBubble) {
          GoodBubble gb=(GoodBubble) b;
          gameObjects.add(gb.sweet);
        }
        break;
      }
    }
  }
}
